﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Scheduling.MRCPSP {

  [Item("MRCPSP TwoPoint Crossover", "Applies two-point crossovers to cross a multi-encoding.")]
  [StorableClass]
  public sealed class MrcpspTwoPointCrossover : SingleSuccessorOperator, ICrossover, IStochasticOperator {

    public MrcpspTwoPointCrossover() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("RandomKeyParents", ""));
      Parameters.Add(new ScopeTreeLookupParameter<IntegerVector>("ModeListParents", ""));
      Parameters.Add(new LookupParameter<RealVector>("RandomKeyChild", ""));
      Parameters.Add(new LookupParameter<IntegerVector>("ModeListChild", ""));
    }

    [StorableConstructor]
    private MrcpspTwoPointCrossover(bool deserializing) : base(deserializing) { }

    private MrcpspTwoPointCrossover(MrcpspTwoPointCrossover original, Cloner cloner) : base(original, cloner) {
    }

    public ILookupParameter<IntegerVector> ModeListChild => (ILookupParameter<IntegerVector>)Parameters["ModeListChild"];

    public IScopeTreeLookupParameter<IntegerVector> ModeListParents => (IScopeTreeLookupParameter<IntegerVector>)Parameters["ModeListParents"];

    public ILookupParameter<RealVector> RandomKeyChild => (ILookupParameter<RealVector>)Parameters["RandomKeyChild"];

    public IScopeTreeLookupParameter<RealVector> RandomKeyParents =>
      (IScopeTreeLookupParameter<RealVector>)Parameters["RandomKeyParents"];

    public ILookupParameter<IRandom> RandomParameter => (ILookupParameter<IRandom>)Parameters["Random"];

    public override IOperation Apply() {
      var random = RandomParameter.ActualValue;
      var length = RandomKeyParents.ActualValue[0].Length;
      var breakpoint1 = random.Next(0, length);
      var breakpoint2 = (breakpoint1 + random.Next(1, length)) % length;
      var breakpoints = new[] { breakpoint1, breakpoint2 };
      Array.Sort(breakpoints);

      RandomKeyChild.ActualValue = Cross(RandomKeyParents.ActualValue, breakpoints);
      ModeListChild.ActualValue = Cross(ModeListParents.ActualValue, breakpoints);

      return base.Apply();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MrcpspTwoPointCrossover(this, cloner);
    }

    private static IntegerVector Cross(ItemArray<IntegerVector> parents, IReadOnlyList<int> breakpoints) {
      var parent1 = parents.First();
      var parent2 = parents.Last();
      var length = parent1.Length;
      var result = new int[length];

      //perform crossover
      var arrayIndex = 0;
      var breakPointIndex = 0;
      var firstParent = true;

      while (arrayIndex < length) {
        if (breakPointIndex < breakpoints.Count &&
            arrayIndex == breakpoints[breakPointIndex]) {
          breakPointIndex++;
          firstParent = !firstParent;
        }

        result[arrayIndex] = firstParent ? parent1[arrayIndex] : parent2[arrayIndex];

        arrayIndex++;
      }

      return new IntegerVector(result);
    }

    private static double Scale(double value, double min, double max, double newMin = 0.0, double newMax = 1.0) =>
      (newMax - newMin) * (value - min) / (max - min) + newMin;

    private static RealVector Scale(RealVector vector) {
      var min = vector.Min();
      var max = vector.Max();

      var scaledVector = new RealVector(vector.Length);
      for (var i = 0; i < vector.Length; i++) {
        scaledVector[i] = Scale(vector[i], min, max);
      }

      return scaledVector;
    }

    private static RealVector Cross(ItemArray<RealVector> parents, IReadOnlyList<int> breakpoints) {
      var parent1 = Scale(parents.First());
      var parent2 = Scale(parents.Last());
      var length = parent1.Length;
      var result = new double[length];

      //perform crossover
      var arrayIndex = 0;
      var breakPointIndex = 0;
      var firstParent = true;

      while (arrayIndex < length) {
        if (breakPointIndex < breakpoints.Count &&
            arrayIndex == breakpoints[breakPointIndex]) {
          breakPointIndex++;
          firstParent = !firstParent;
        }

        result[arrayIndex] = firstParent ? parent1[arrayIndex] : parent2[arrayIndex];

        arrayIndex++;
      }

      return new RealVector(result);
    }
  }
}
